/*
 * Connect Four
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of Connect Four.
 *
 * Connect Four is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Connect Four is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;
using System.Reflection;

namespace ConnectFour {
  public partial class ConnectFourForm : Form {

    [StructLayout(LayoutKind.Sequential)]
    public struct PeekMsg {
      public IntPtr hWnd;
      public Message msg;
      public IntPtr wParam;
      public IntPtr lParam;
      public uint time;
      public System.Drawing.Point p;
    }

    [System.Security.SuppressUnmanagedCodeSecurity]
    [DllImport("User32.dll", CharSet = CharSet.Auto)]
    public static extern bool PeekMessage(out PeekMsg msg, IntPtr hWnd,
             uint messageFilterMin, uint messageFilterMax, uint flags);

    const int WIDTH = 448;
    const int HEIGHT = 448;
    const int FRAMES_PER_SECOND = 20;

    public long FREQUENCY = Stopwatch.Frequency;
    public long TICKS_PER_FRAME = Stopwatch.Frequency / FRAMES_PER_SECOND;
    private Bitmap bitmap;
    private Graphics g;
    public Graphics graphics;
    public Stopwatch stopwatch = new Stopwatch();
    public long nextFrameStart;
    private Bitmap backgroundBitmap;
    private Bitmap blackCheckerBitmap;
    private Bitmap boardBitmap;
    private Bitmap greenCheckerBitmap;
    private Bitmap redCheckerBitmap;
    private bool animating;
    private CellState[,] cellStates = new CellState[6, 7];
    private int playerPosition = 3;
    private int nextPlayerPosition = 3;
    private double playerPieceX = 192;
    private double playerPieceVX;
    private int playerPieceMoveCount;
    private GameState gameState = GameState.UserChoosing;
    private CellState droppingCell;
    private double droppingX;
    private double droppingY;
    private double droppingTargetY;
    private double droppingVY;
    private int droppingCompleteRow;
    private int droppingCompleteColumn;
    private Minimax minimax = new Minimax();
    private MinimaxResultMapAvailable minimaxResultMapAvailable;
    private MinimaxBlackMoveAvailable minimaxBlackMoveAvailable;
    private bool resultsAvailable;
    private int[,] results;
    private int[,] win = new int[4, 2];
    private bool redWon;
    private bool blinkOn;
    private int blinkOnCount;
    private int blinkingCount;
    private List<DroppingPiece> droppingPieces = new List<DroppingPiece>();
    private Random random = new Random();
    private bool redPlayedLast = true;
    private int maxDepth = 1;
    private Icon redIcon;
    private Icon blackIcon;
    private Font levelFont;
    private int levelX;
    private int levelY;
    private int levelXCenter;
    private int levelDelay;

    public ConnectFourForm() {
      InitializeComponent();
      ClientSize = new Size(WIDTH, HEIGHT);
      StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
      SetStyle(ControlStyles.UserPaint, true);
      SetStyle(ControlStyles.AllPaintingInWmPaint, true);
      SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
    }

    private void ConnectFourForm_Load(object sender, EventArgs e) {
      minimaxResultMapAvailable
          = new MinimaxResultMapAvailable(ProcessMinimaxResultMap);
      minimaxBlackMoveAvailable
          = new MinimaxBlackMoveAvailable(ProcessMinimaxBlackMove);
      minimax.MinimaxBlackMoveAvailableEvent += minimaxBlackMoveAvailable;
      minimax.MinimaxResultMapAvailableEvent += minimaxResultMapAvailable;
      bitmap = new Bitmap(WIDTH, HEIGHT, graphics = CreateGraphics());
      g = Graphics.FromImage(bitmap);
      g.SmoothingMode = SmoothingMode.None;
      LoadImages();
      ResetCellStates();
      minimax.MaxDepth = maxDepth;
      minimax.Reset();
      minimax.ConsiderAllPossibleRedMoves();
      Text = "Connect Four - Level " + maxDepth;
      Icon = redIcon;
      levelFont = new Font(Font.FontFamily, 60, FontStyle.Bold);
      levelY = (int)((448 - levelFont.Height) / 2);
      stopwatch.Start();
      nextFrameStart = stopwatch.ElapsedTicks;
      Application.Idle += OnApplicationIdle;
    }

    private void OnApplicationIdle(object sender, EventArgs e) {
      if (animating) {
        PeekMsg msg;
        while (!PeekMessage(out msg, IntPtr.Zero, 0, 0, 0)) {
          do {
            UpdateModel();
            nextFrameStart += TICKS_PER_FRAME;
          } while (nextFrameStart < stopwatch.ElapsedTicks);
          RenderFrame(graphics);
          long remainingTicks = nextFrameStart - stopwatch.ElapsedTicks;
          if (remainingTicks > 0) {
            Thread.Sleep((int)(1000 * remainingTicks / FREQUENCY));
          }
        }
      }
    }

    private void UpdateModel() {
      switch(gameState) {
        case GameState.LevelEntering: {
          int nextLevelX = levelX + 48;
          if (nextLevelX >= levelXCenter) {
            levelX = levelXCenter;
            levelDelay = 25;
            GameState = GameState.LevelWaiting;
          } else {
            levelX = nextLevelX;
          }
          break;
        }
        case GameState.LevelWaiting:
          if (levelDelay == 0) {
            GameState = GameState.LevelExiting;
          } else {
            levelDelay--;
          }
          break;
        case GameState.LevelExiting:
          levelX += 48;
          if (levelX > 448) {
            EndLevelExiting();
          }
          break;
        case GameState.UserChoosing:
          if (playerPieceMoveCount > 0) {        
            playerPieceX += playerPieceVX;
            playerPieceMoveCount--;
            if (playerPieceMoveCount == 0) {
              playerPieceX = playerPosition << 6;
              StopAnimating();
            }
          }
          break;
        case GameState.Dropping:
          if (droppingY + droppingVY < droppingTargetY) {
            droppingVY += 6;
            droppingY += droppingVY;
          } else {
            cellStates[droppingCompleteRow, droppingCompleteColumn] = droppingCell;
            StopAnimating();            
            if (droppingCell == CellState.Black) {
              playerPosition = nextPlayerPosition;
              GameState = GameState.UserChoosing;
              BoardState boardState;
              minimax.PositionCell(droppingCompleteRow,
                  droppingCompleteColumn, CellState.Black, out boardState, win);
              if (boardState != BoardState.Intermediate) {
                if (boardState == BoardState.Stalemate) {
                  StartClearing();
                } else {
                  StartBlinking(false);
                }
              } else {
                minimax.ConsiderAllPossibleRedMoves();
              }
            } else {
              if (resultsAvailable) {
                resultsAvailable = false;
                BoardState boardState;
                minimax.PositionCell(droppingCompleteRow,
                    droppingCompleteColumn, CellState.Red, out boardState, win);
                if (boardState != BoardState.Intermediate) {
                  if (boardState == BoardState.Stalemate) {
                    StartClearing();
                  } else {
                    StartBlinking(true);
                  }
                } else {
                  DropBlack(results[playerPosition, 0], results[playerPosition, 1]);
                }
              } else {
                GameState = GameState.ComputerChoosing;
                minimax.RedMoved(droppingCompleteRow, droppingCompleteColumn);                
              }              
            }
          }
          break;
        case GameState.Blinking:
          if (blinkingCount > 0) {
            blinkingCount--;
            if (blinkOnCount > 0) {
              blinkOnCount--;
            } else {
              blinkOn = !blinkOn;
              blinkOnCount = FRAMES_PER_SECOND >> 1;
            }
          } else {
            blinkOn = false;
            StopAnimating();
            StartClearing();
          }
          break;
        case GameState.Clearing:
          bool pieceVisible = false;
          foreach (DroppingPiece piece in droppingPieces) {
            //Console.WriteLine(piece.y);
            if (piece.pauseCount > 0) {
              piece.pauseCount--;
              pieceVisible = true;
            } else {
              piece.vy += 6;
              piece.y += piece.vy;              
              if (piece.y < 448) {
                pieceVisible = true;
              }
            }
          }
          if (!pieceVisible) {
            EndClearing();
          }
          break;
      }
    }

    private GameState GameState {
      set {
        gameState = value;
        if (gameState == GameState.ComputerChoosing) {
          Icon = blackIcon;
          Cursor = Cursors.WaitCursor;
        } else {
          Cursor = Cursors.Default;
        }
        if (gameState == GameState.UserChoosing) {
          Icon = redIcon;
        } else if (gameState == GameState.Dropping) {
          if (droppingCell == CellState.Red) {
            Icon = redIcon;
          } else {
            Icon = blackIcon;
          }
        } else if (gameState == GameState.Blinking) {
          if (redWon) {
            Icon = redIcon;
          } else {
            Icon = blackIcon;
          }
        }
      }
      get {
        return gameState;
      }
    }

    private void DropBlack(int row, int column) {
      droppingCell = CellState.Black;
      GameState = GameState.Dropping;      
      droppingX = column << 6;
      droppingY = 0;
      droppingVY = 0;
      droppingTargetY = 64 + (row << 6);
      droppingCompleteRow = row;
      droppingCompleteColumn = column;
      StartAnimating();
    }

    private void RenderFrame(Graphics graphics) {
      g.DrawImageUnscaled(backgroundBitmap, 0, 0);
      if (GameState == GameState.UserChoosing) {
        g.DrawImageUnscaled(redCheckerBitmap, (int)playerPieceX, 0);
      } else if (GameState == GameState.Dropping) {
        g.DrawImageUnscaled(
            droppingCell == CellState.Red ? redCheckerBitmap : blackCheckerBitmap, 
            (int)droppingX, (int)droppingY);
      }
      if (GameState == GameState.Clearing) {
        foreach(DroppingPiece piece in droppingPieces) {
          g.DrawImageUnscaled(
              piece.cellState == CellState.Red ? redCheckerBitmap : blackCheckerBitmap, 
              piece.x, piece.y);
        }
      } else {
        for (int i = 0, y = 64; i < 6; i++, y += 64) {
          for (int j = 0, x = 0; j < 7; j++, x += 64) {
            switch (cellStates[i, j]) {
              case CellState.Red:
                g.DrawImageUnscaled(redCheckerBitmap, x, y);
                break;
              case CellState.Black:
                g.DrawImageUnscaled(blackCheckerBitmap, x, y);
                break;
            }
          }
        }
        if (blinkOn) {
          for (int i = 0; i < 4; i++) {
            g.DrawImageUnscaled(greenCheckerBitmap, win[i, 1] << 6,
                64 + (win[i, 0] << 6));
          }
        }
      }
      g.DrawImageUnscaled(boardBitmap, 0, 0);
      if (GameState == GameState.LevelEntering
          || GameState == GameState.LevelWaiting
          || GameState == GameState.LevelExiting) {
        g.DrawString("LEVEL " + maxDepth, levelFont,
            Brushes.DarkViolet, levelX, levelY);
      }      
      graphics.DrawImageUnscaled(bitmap, 0, 0);
    }

    private void LoadImages() {
      Assembly assembly = Assembly.GetExecutingAssembly();
      redIcon = new Icon(assembly.GetManifestResourceStream(
          "ConnectFour.images.redchecker.ico"));
      blackIcon = new Icon(assembly.GetManifestResourceStream(
          "ConnectFour.images.blackchecker.ico"));
      backgroundBitmap = new Bitmap(assembly.GetManifestResourceStream(
          "ConnectFour.images.background.jpg"));
      blackCheckerBitmap = new Bitmap(assembly.GetManifestResourceStream(
          "ConnectFour.images.blackchecker.gif"));
      Bitmap cellBitmap = new Bitmap(assembly.GetManifestResourceStream(
          "ConnectFour.images.board.gif"));      
      greenCheckerBitmap = new Bitmap(assembly.GetManifestResourceStream(
          "ConnectFour.images.greenchecker.gif"));
      redCheckerBitmap = new Bitmap(assembly.GetManifestResourceStream(
          "ConnectFour.images.redchecker.gif"));

      boardBitmap = new Bitmap(WIDTH, HEIGHT, graphics);
      using (Graphics g = Graphics.FromImage(boardBitmap)) {
        for (int i = 0, y = 64; i < 6; i++, y += 64) {
          for (int j = 0, x = 0; j < 7; j++, x += 64) {
            g.DrawImageUnscaled(cellBitmap, x, y);
          }
        }
      }
    }

    private void ConnectFourForm_Paint(object sender, PaintEventArgs e) {
      RenderFrame(e.Graphics);
    }

    private void ResetCellStates() {
      for(int i = 0; i < 6; i++) {
        for(int j = 0; j < 7; j++) {
          cellStates[i, j] = CellState.Empty;
        }
      }
    }

    private void ConnectFourForm_MouseMove(object sender, MouseEventArgs e) {
      if (e.X < 0 || e.Y < 0 || e.X >= 448 || e.Y >= 488) {
        return;
      }
      if (GameState == GameState.UserChoosing) {
        int newPlayerPosition = e.X >> 6;
        if (newPlayerPosition != playerPosition) {
          playerPieceMoveCount = 3;
          playerPieceVX = ((newPlayerPosition << 6) - playerPieceX) / 3.0;
          playerPosition = newPlayerPosition;
          StartAnimating();
        }
      } else {
        nextPlayerPosition = e.X >> 6;
        playerPieceX = nextPlayerPosition << 6;
      }
    }

    private void ConnectFourForm_MouseClick(object sender, MouseEventArgs e) {
      if (e.X < 0 || e.Y < 0 || e.X >= 448 || e.Y >= 488) {
        return;
      }
      if (GameState == GameState.UserChoosing) {
        for (int i = 5; i >= 0; i--) {
          if (cellStates[i, playerPosition] == CellState.Empty) {
            droppingCell = CellState.Red;
            droppingX = playerPosition << 6;
            droppingY = 0;
            droppingVY = 0;
            droppingTargetY = 64 + (i << 6);
            droppingCompleteRow = i;
            droppingCompleteColumn = playerPosition;
            GameState = GameState.Dropping;
            StartAnimating();
            break;
          }
        }
      }
    }

    private void StartAnimating() {
      if (!animating) {
        animating = true;
        nextFrameStart = stopwatch.ElapsedTicks;
      }
    }

    private void StopAnimating() {
      if (animating) {
        animating = false;
      }
    }

    private void ProcessMinimaxResultMap(int[,] results) {
      if (InvokeRequired) {
        Invoke(minimaxResultMapAvailable, results);
        return;
      }

      resultsAvailable = true;
      this.results = results;

      /*for (int i = 0; i < 7; i++) {
        Console.WriteLine("{0} -> {1},{2}", i, results[i, 0], results[i, 1]);
      }*/

      if (GameState == GameState.ComputerChoosing) {
        Cursor = Cursors.Default;
        resultsAvailable = false;
        BoardState boardState;
        minimax.PositionCell(droppingCompleteRow,
            droppingCompleteColumn, CellState.Red, out boardState, win);
        if (boardState != BoardState.Intermediate) {
          if (boardState == BoardState.Stalemate) {
            StartClearing();
          } else {
            StartBlinking(true);
          }
        } else {
          DropBlack(results[playerPosition, 0], results[playerPosition, 1]);
        }
      }
    }

    private void ProcessMinimaxBlackMove(int row, int column) {
      if (InvokeRequired) {
        Invoke(minimaxBlackMoveAvailable, row, column);
        return;
      }
      DropBlack(row, column);
    }

    private void StartLevelEntering() {
      GameState = GameState.LevelEntering;
      StartAnimating();
      
      int width = (int)(g.MeasureString("LEVEL " + maxDepth, levelFont).Width);
      levelX = -width;
      levelXCenter = (448 - width) / 2;
    }

    private void StartBlinking(bool redWon) {
      this.redWon = redWon;
      GameState = GameState.Blinking;
      blinkOn = true;
      blinkingCount = FRAMES_PER_SECOND * 4;
      blinkOnCount = FRAMES_PER_SECOND >> 1;
      StartAnimating();
    }

    private void StartClearing() {
      StartAnimating();
      GameState = GameState.Clearing;
      droppingPieces.Clear();
      for (int j = 0; j < 7; j++) {
        int pauseCount = 8 + random.Next(8);
        for (int i = 0; i < 6; i++) {          
          if (cellStates[i, j] != CellState.Empty) {
            DroppingPiece piece = new DroppingPiece();
            piece.x = j << 6;
            piece.y = 64 + (i << 6);
            piece.vy = 0;
            piece.pauseCount = pauseCount;
            piece.cellState = cellStates[i, j];
            droppingPieces.Add(piece);
          }
        }
      }
    }

    private void EndLevelExiting() {
      StopAnimating();
      GameState = GameState.ComputerChoosing;
      minimax.RequestBlackMove();
    }

    private void EndClearing() {
      StopAnimating();
      ResetCellStates();
      if (redWon) {
        maxDepth++;
      }
      minimax.MaxDepth = maxDepth;
      minimax.Reset();
      Text = "Connect Four - Level " + maxDepth;
      if (redWon) {
        redWon = false;
        StartLevelEntering();
        redPlayedLast = false;        
      } else {
        if (redPlayedLast) {
          GameState = GameState.ComputerChoosing;
          minimax.RequestBlackMove();
        } else {
          playerPosition = nextPlayerPosition;
          GameState = GameState.UserChoosing;
          minimax.ConsiderAllPossibleRedMoves();
        }
        redPlayedLast = !redPlayedLast;
      }
    }
  }

  public enum CellState { Empty, Red, Black };

  public enum GameState { 
    LevelEntering, LevelWaiting, LevelExiting, 
    UserChoosing, Dropping, ComputerChoosing, Blinking, Clearing 
  };

  class DroppingPiece {
    public int pauseCount;
    public int x;
    public int y;
    public int vy;
    public CellState cellState;
  }
}