/*
 * Connect Four
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of Connect Four.
 *
 * Connect Four is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Connect Four is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace ConnectFour {

  public class Minimax {

    private int[, ,] scans = {
      { { 0, 1 }, { 0, 2 }, { 0, 3 } },
      { { 1, 1 }, { 2, 2 }, { 3, 3 } },
      { { 1, 0 }, { 2, 0 }, { 3, 0 } },
      { { 1, -1 }, { 2, -2 }, { 3, -3 } },
      { { 0, -1 }, { 0, -2 }, { 0, -3 } },
      { { -1, -1 }, { -2, -2 }, { -3, -3 } },
      { { -1, 0 }, { -2, 0 }, { -3, 0 } },
      { { -1, 1 }, { -2, 2 }, { -3, 3 } },

      { { 0, -1 }, { 0, 1 }, { 0, 2 } },
      { { -1, -1 }, { 1, 1 }, { 2, 2 } },
      { { -1, 0 }, { 1, 0 }, { 2, 0 } },
      { { -1, 1 }, { 1, -1 }, { 2, -2 } },
      { { 0, 1 }, { 0, -1 }, { 0, -2 } },
      { { 1, 1 }, { -1, -1 }, { -2, -2 } },
      { { 1, 0 }, { -1, 0 }, { -2, 0 } },
      { { 1, -1 }, { -1, 1 }, { -2, 2 } },
    };

    private int[,] waysToWin = new int[6, 7];
    private Random random = new Random();
    private Examination[][] examinations = new Examination[43][];
    public event MinimaxResultMapAvailable MinimaxResultMapAvailableEvent;
    public event MinimaxBlackMoveAvailable MinimaxBlackMoveAvailableEvent;
    private object redMoveMonitor = new Object();
    private int redMoveColumn;
    private int redMoveRow;
    private bool redMoved;
    private CellState[,] board = new CellState[6, 7];
    private int[,] results = new int[7, 2];
    private int[] frees = new int[7];
    private int maxDepth;

    public Minimax() {
      for(int i = 0; i < 43; i++) {
        examinations[i] = new Examination[7];
        for(int j = 0; j < 7; j++) {
          examinations[i][j] = new Examination();
        }
      }
      for (int i = 0; i < 6; i++) {
        for (int j = 0; j < 7; j++) {
          for (int k = 0; k < 16; k++) {
            for (int m = 0; m < 3; m++) {
              if (!ValidPosition(i + scans[k, m, 0], j + scans[k, m, 1])) {
                goto nextScan;
              }
            }
            waysToWin[i, j]++;
            nextScan:;
          }
        }
      }
    }

    public void Reset() {
      for (int j = 0; j < 7; j++) {
        for (int i = 0; i < 6; i++) {
          frees[j] = 5;
          board[i, j] = CellState.Empty;
        }
      }
    }

    public bool ValidPosition(int row, int column) {
      return row >= 0 && column >= 0 && row < 6 && column < 7;
    }

    public void RequestBlackMove() {
      ThreadPool.QueueUserWorkItem(new WaitCallback(MakeBlackMove));
    }

    private void MakeBlackMove(Object stateInfo) {
      int bestRow;
      int bestColumn;
      int bestValue;
      BoardState bestState;
      int bestDepth;
      ComputeMinimax(board, frees, false, 0, maxDepth, -10000, 10000, 0,
          out bestRow, out bestColumn, out bestValue, out bestState, out bestDepth);
      MinimaxBlackMoveAvailableEvent(bestRow, bestColumn);
    }

    public void ConsiderAllPossibleRedMoves() {
      //Console.WriteLine("Considering all possible red moves...");
      ThreadPool.QueueUserWorkItem(new WaitCallback(MakeAllPossibleRedMoves));
    }

    public void MakeAllPossibleRedMoves(Object stateInfo) {

      //Thread.Sleep(3000);

      // Compute estimated value
      int estimatedValue = 0;
      for (int i = 0; i < 6; i++) {
        for (int j = 0; j < 7; j++) {
          switch (board[i, j]) {
            case CellState.Black:
              estimatedValue -= waysToWin[i, j];
              break;
            case CellState.Red:
              estimatedValue += waysToWin[i, j];
              break;
          }
        }
      }
      
      // Compute examination path
      Examination[] exams = examinations[42];
      int examsCount = 0;
      for (int i = 0; i < 7; i++) {
        Examination examination = exams[i];
        examination.row = examination.column = -1;
        examination.maxPlaying = true;
        results[i, 0] = results[i, 1] = -1;
      }
      for (int i = 0; i < 7; i++) {                
        if (frees[i] >= 0) {
          Examination examination = exams[examsCount];
          examination.estimateValue = estimatedValue + waysToWin[frees[i], i];
          examination.row = frees[i];
          examination.column = i;
          examsCount++;
        } 
      }
      if (examsCount > 1) {
        for (int i = 0; i < 25; i++) {
          int index0 = random.Next(examsCount);
          int index1 = random.Next(examsCount);
          Examination temp = exams[index0];
          exams[index0] = exams[index1];
          exams[index1] = temp;
        }
      }
      Array.Sort<Examination>(exams, 0, examsCount);

      // Test out all possible red moves
      for (int i = 0; i < examsCount; i++) {
        int column = exams[i].column;
        board[exams[i].row, column] = CellState.Red;        
        frees[column]--;
        int bestRow;
        int bestColumn;
        int bestValue;
        BoardState bestState;
        int bestDepth;
        //Console.WriteLine("Considering column: {0}", column);
        ComputeMinimax(board, frees, false, 0, maxDepth, -10000, 10000, 
            exams[i].estimateValue,
            out bestRow, out bestColumn, out bestValue, out bestState, out bestDepth);
        results[column, 0] = bestRow;
        results[column, 1] = bestColumn;
        //Console.WriteLine("{0} => {1},{2},{3},{4}",
        //    column, bestRow, bestColumn, bestState, bestValue);
        frees[column]++;
        board[exams[i].row, column] = CellState.Empty;        
      }
      lock (redMoveMonitor) {        
        if (redMoved) {
          board[redMoveRow, redMoveColumn] = CellState.Red;          
          redMoved = false;
        } 
        MinimaxResultMapAvailableEvent(results);
      }      
    }

    public void PositionCell(int row, int column, CellState cellState,
        out BoardState boardState, int[,] win) {      
      frees[column]--;
      board[row, column] = cellState;

      for (int j = 0; j < 16; j++) {
        if (ValidPosition(row + scans[j, 0, 0], column + scans[j, 0, 1])
            && ValidPosition(row + scans[j, 1, 0], column + scans[j, 1, 1])
            && ValidPosition(row + scans[j, 2, 0], column + scans[j, 2, 1])
            && board[row + scans[j, 0, 0], column + scans[j, 0, 1]]
                == cellState
            && board[row + scans[j, 1, 0], column + scans[j, 1, 1]]
                == cellState
            && board[row + scans[j, 2, 0], column + scans[j, 2, 1]]
                == cellState) {
          win[0, 0] = row;
          win[0, 1] = column;
          for (int k = 0; k < 3; k++) {
            win[k + 1, 0] = row + scans[j, k, 0];
            win[k + 1, 1] = column + scans[j, k, 1];
          }
          boardState = cellState == CellState.Red 
              ? BoardState.MaxWins : BoardState.MinWins;
          return;
        }
      }

      for (int j = 0; j < 7; j++) {
        if (frees[j] >= 0) {
          goto notStalemate;
        }
      }
      boardState = BoardState.Stalemate;
      return;

      notStalemate:
      
      boardState = BoardState.Intermediate;
    }

    public int MaxDepth {
      set {
        maxDepth = value;
      }
      get {
        return maxDepth;
      }
    }

    public void RedMoved(int row, int column) {
      lock (redMoveMonitor) {
        redMoveRow = row;
        redMoveColumn = column;
        redMoved = true;
      }
    }

    private void ComputeMinimax(CellState[,] board, int[] frees, bool maxPlaying, 
        int depth, int maxDepth, int bestMaxValue, int bestMinValue,
        int estimatedValue,
        out int bestRow, out int bestColumn, 
        out int bestValue, out BoardState bestState,
        out int bestDepth) {

      lock (redMoveMonitor) {
        if (redMoved) {
          if (board[redMoveRow, redMoveColumn] != CellState.Red) {
            bestDepth = bestRow = bestColumn = bestValue = -1;
            bestState = BoardState.Intermediate;
            return;
          }
        }
      }

      CellState player = maxPlaying ? CellState.Red : CellState.Black;

      // Compute examination path
      Examination[] exams = examinations[depth];
      int examsCount = 0;
      for (int i = 0; i < 7; i++) {
        Examination examination = exams[i];
        examination.row = examination.column = -1;
        examination.maxPlaying = maxPlaying;
      }
      for (int i = 0; i < 7; i++) {
        if (frees[i] >= 0) {
          Examination examination = exams[examsCount];
          examination.estimateValue = estimatedValue 
              + (maxPlaying ? waysToWin[frees[i], i] : -waysToWin[frees[i], i]);
          examination.row = frees[i];
          examination.column = i;
          examsCount++;
        }
      }      
      if (examsCount > 1) {
        for (int i = 0; i < 25; i++) {
          int index0 = random.Next(examsCount);
          int index1 = random.Next(examsCount);
          Examination temp = exams[index0];
          exams[index0] = exams[index1];
          exams[index1] = temp;
        }
      }
      Array.Sort<Examination>(exams, 0, examsCount);

      // Examine possible moves
      if (maxPlaying) {
        bestState = BoardState.MinWins;
        bestValue = -10000;
      } else {
        bestState = BoardState.MaxWins;
        bestValue = 10000;
      }
      bestRow = -1;
      bestColumn = -1;
      bestDepth = -1;
      for (int i = 0; i < examsCount; i++) {
        Examination exam = exams[i];
        frees[exam.column]--;
        board[exam.row, exam.column] = player;
        int value = exam.estimateValue;
        int depthFound = depth;
        BoardState state = BoardState.Intermediate;

        // Test if player won
        bool terminateSearch = false;
        for (int j = 0; j < 16; j++) {
          if (ValidPosition(exam.row + scans[j, 0, 0], exam.column + scans[j, 0, 1])
              && ValidPosition(exam.row + scans[j, 1, 0], exam.column + scans[j, 1, 1])
              && ValidPosition(exam.row + scans[j, 2, 0], exam.column + scans[j, 2, 1])
              && board[exam.row + scans[j, 0, 0], exam.column + scans[j, 0, 1]] 
                  == player
              && board[exam.row + scans[j, 1, 0], exam.column + scans[j, 1, 1]] 
                  == player
              && board[exam.row + scans[j, 2, 0], exam.column + scans[j, 2, 1]] 
                  == player) {
            terminateSearch = true;
            bestColumn = exam.column;
            bestRow = exam.row;
            state = bestState = maxPlaying ? BoardState.MaxWins : BoardState.MinWins;
            value = bestValue = maxPlaying ? 1000 : -1000;          
            goto loopEnd;
          }
        }

        // Test for stalemate
        bool isStalemate = false;
        for (int j = 0; j < 7; j++) {
          if (frees[j] >= 0) {
            goto notStalemate;
          }
        }
        value = 0;
        state = BoardState.Stalemate;
        isStalemate = true;

        notStalemate:

        if (maxPlaying) {
          if (!isStalemate && depth < maxDepth) {
            int row, col;
            ComputeMinimax(board, frees, !maxPlaying, depth + 1, maxDepth, 
                Math.Max(bestMaxValue, bestValue),
                bestMinValue, value, out row, out col, out value, out state, 
                out depthFound);
          }
          if (value > bestValue
              || (value == bestValue && value == -1000 && depthFound > bestDepth)) {
            bestValue = value;
            bestColumn = exam.column;
            bestRow = exam.row;
            bestState = state;
            bestDepth = depthFound;
          }
          if (bestValue >= bestMinValue) {
            terminateSearch = true;
          }
        } else {
          if (!isStalemate && depth < maxDepth) {
            int row, col;
            ComputeMinimax(board, frees, !maxPlaying, depth + 1, maxDepth, bestMaxValue,
                Math.Min(bestMinValue, bestValue), value, out row, out col, 
                out value, out state, out depthFound);
          }
          if (value < bestValue 
              || (value == bestValue && value == 1000 && depthFound > bestDepth)) {
            bestValue = value;
            bestColumn = exam.column;
            bestRow = exam.row;
            bestState = state;
            bestDepth = depthFound;
          }
          if (bestValue <= bestMaxValue) {
            terminateSearch = true;
          }
        }

        loopEnd:

        frees[exam.column]++;
        board[exam.row, exam.column] = CellState.Empty;

        if (terminateSearch) {
          break;
        }
      }
    }
  }

  class Examination : IComparable<Examination> {
    public int row;
    public int column;
    public int estimateValue;
    public bool maxPlaying;

    #region IComparable<Examination> Members

    public int CompareTo(Examination other) {
      if (maxPlaying) {
        if (estimateValue < other.estimateValue) {
          return 1;
        } else if (estimateValue > other.estimateValue) {
          return -1;
        } else {
          return 0;
        }
      } else {
        if (estimateValue < other.estimateValue) {
          return -1;
        } else if (estimateValue > other.estimateValue) {
          return 1;
        } else {
          return 0;
        }
      }
    }

    #endregion
  }

  public delegate void MinimaxResultMapAvailable(int[,] results);
  public delegate void MinimaxBlackMoveAvailable(int row, int column);

  public enum BoardState { Intermediate, Stalemate, MaxWins, MinWins };
}
